
#include "loadlib.h"

/* ==================================================== */
typedef void (*INITFUN) (lua_State*);
/* ==================================================== */
int ldlib_loadlib    ( lua_State *L );
int ldlib_unloadlib  ( lua_State *L );
int ldlib_gclib      ( lua_State *L );
/* ==================================================== */
void lua_loadlibopen( lua_State *L ) {
   int tag_hlib,tag_closedlib;

   tag_hlib = lua_newtag(L);
   tag_closedlib = lua_newtag(L);

   lua_newtable(L); lua_setglobal(L,"_LOADLIB");

   lua_pushnumber(L,tag_hlib); lua_pushnumber(L,tag_closedlib);
   lua_pushcclosure(L,ldlib_loadlib,2); lua_setglobal(L,"loadlib");

   lua_pushnumber(L,tag_hlib); lua_pushnumber(L,tag_closedlib);
   lua_pushcclosure(L,ldlib_unloadlib,2); lua_setglobal(L,"unloadlib");

   lua_pushcclosure(L,ldlib_gclib,0); lua_settagmethod(L,tag_hlib,"gc");

}
/* ==================================================== */
int ldlib_loadlib( lua_State *L ) {
   char errmsg[256]; int n;
   HMODULE hlib; INITFUN fun;
   const char *slib, *sfun, *x=NULL;
   int tag_hlib = (int) lua_tonumber(L,-2); lua_pop(L,2);
   
   n=lua_gettop(L);
   slib = lua_tostring(L,1); 
   sfun = lua_tostring(L,2); 
   if(n==3) x = lua_tostring(L,3); lua_pop(L,n);

   lua_getglobal(L,"_LOADLIB");
   lua_pushstring(L,slib); lua_rawget(L,1);
   
   if( lua_isnil(L,2) ) {
      hlib = DLOPEN( slib );
      
      if( hlib==NULL ) { 
         _snprintf( errmsg, sizeof(errmsg)-1, "Cannot load plug-in library %s.", slib ); errmsg[sizeof(errmsg)-1]=0;
         if( x && *x=='x' ) { lua_settop(L,0); lua_pushnil(L); lua_pushstring(L,errmsg); return 2; }
         lua_error(L,errmsg);
         return 0;
      }

      lua_pop(L,1); lua_pushstring(L,slib); lua_pushusertag(L,hlib,tag_hlib);
      lua_rawset(L,1);
   }
   else {
      hlib = (HMODULE)lua_touserdata(L,2);
      lua_pop(L,1);
   }

   fun = (INITFUN) DLSYM(hlib,sfun);
   if( fun==NULL ) { 
      _snprintf( errmsg, sizeof(errmsg)-1, "Cannot locate function %s within plug-in library %s.", sfun, slib ); errmsg[sizeof(errmsg)-1]=0;
      if( x && *x=='x' ) { lua_settop(L,0); lua_pushnil(L); lua_pushstring(L,errmsg); return 2; }
      lua_error(L,errmsg);
      return 0;
   }

   fun(L); lua_settop(L,0); lua_pushnumber(L,1); return 1;
}
/* ==================================================== */
int ldlib_unloadlib( lua_State *L ) {
   HMODULE hlib; const char *slib;
   int tag_hlib = (int) lua_tonumber(L,-2);
   int tag_closedlib = (int) lua_tonumber(L,-1); lua_pop(L,2);

   slib = lua_tostring(L,1); lua_pop(L,1);

   lua_getglobal(L,"_LOADLIB" );
   lua_pushstring(L,slib); lua_rawget(L,1);

   if( lua_isnil(L,2) ) lua_error(L,"library not loaded" );

   if( lua_tag(L,2) != tag_hlib ) lua_error(L,"cannot unload non-library object" );

   hlib = (HMODULE)lua_touserdata(L,2);
   DLCLOSE(hlib); lua_settag(L,tag_closedlib);

   lua_pop(L,1); lua_pushstring(L,slib); lua_pushnil(L); lua_rawset(L,1);

   lua_settop(L,0); return 0;
}
/* ==================================================== */
int ldlib_gclib( lua_State *L ) {

   HMODULE hlib = (HMODULE)lua_touserdata(L,1);
   DLCLOSE(hlib);

   lua_settop(L,0); return 0;
}
/* ==================================================== */
